'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var React = require('react');
var reactDom = require('react-dom');
var hoistNonReactStatics = _interopDefault(require('hoist-non-react-statics'));

var DataContext = React.createContext(null);

// `http://localhost:3000/some-rest-api/99`;
// => ["http:", "localhost:3000", "some-rest-api", "99"]
// `http://localhost:3000/some-rest-api/99?foo=bar&baz=lol`;
// => ["http:", "localhost:3000", "some-rest-api", "99?foo=bar&baz=lol"]
var getCacheKeys = function (url) {
    return url.split('/').filter(Boolean);
};

var addToCache = function (cache, url, data) {
    var temp = cache;
    var keys = getCacheKeys(url);
    var lastIndex = keys.length - 1;
    keys.forEach(function (k, index) {
        if (index !== lastIndex) {
            // not the last key
            if (!temp[k])
                temp[k] = {};
            temp = temp[k];
        }
        else {
            // add the data
            temp[k] = temp[k] || {};
            temp[k].__raw = data;
        }
    });
    return cache;
};

var retrieveFromCache = function (cache, url) {
    var temp = cache;
    var keys = getCacheKeys(url);
    var tempKey;
    while (tempKey = keys.shift()) {
        if (!temp[tempKey])
            return undefined;
        temp = temp[tempKey];
    }
    return temp.__raw;
};

var DataProvider = function (_a) {
    var children = _a.children, client = _a.client;
    var addToBePrefetched = React.useCallback(function (url) {
        client.toBePrefetched[url] = true;
    }, [client.toBePrefetched]);
    var addToCache$1 = React.useCallback(function (key, value) {
        addToCache(client.cache, key, value);
        if (!client.ssr) {
            // force an update
            // notify all subscribers that listens for the same key
            // this will trigger an update/re-render for all affected hook
            client.notifySubscribers(key);
        }
    }, [client]);
    var retrieveFromCache$1 = React.useCallback(function (url) {
        return retrieveFromCache(client.cache, url);
    }, [client.cache]);
    var injectedValues = {
        client: client,
        addToCache: addToCache$1,
        addToBePrefetched: addToBePrefetched,
        retrieveFromCache: retrieveFromCache$1,
        fetcher: fetch,
    };
    return (React.createElement(DataContext.Provider, { value: injectedValues }, children));
};

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */

var __assign = function() {
    __assign = Object.assign || function __assign(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};

// returns an array of FetchOptions and string of fetchPolicy
var createFetchOptions = function (fetchOptions, client, dataOpts, lazy) {
    if (lazy === void 0) { lazy = false; }
    var finalMethod = fetchOptions.method && lazy ? fetchOptions.method : 'GET';
    var fetchPolicy = dataOpts.fetchPolicy !== undefined ? dataOpts.fetchPolicy : client.fetchPolicy;
    var ssrOpt = dataOpts.ssr !== undefined ? dataOpts.ssr : true;
    var isSSR = client.ssr && ssrOpt && typeof window === 'undefined';
    if (finalMethod !== 'GET')
        fetchPolicy = 'network-only';
    if (isSSR || client.test)
        fetchPolicy = 'cache-first';
    return [
        __assign(__assign({}, fetchOptions), { 
            // only allow non-GET request for lazy requests, because non-GET request can be not idempotent
            method: finalMethod, headers: __assign(__assign({}, client.headers), fetchOptions.headers) }),
        fetchPolicy,
    ];
};

/* istanbul ignore file */

var has = Object.prototype.hasOwnProperty
  , undef;

/**
 * Transform a query string to an object.
 *
 * @param {Object} obj Object that should be transformed.
 * @param {String} prefix Optional prefix.
 * @returns {String}
 * @api public
 */
function querystringify(obj, prefix) {
  prefix = prefix || '';

  var pairs = []
    , value
    , key;

  //
  // Optionally prefix with a '?' if needed
  //
  if ('string' !== typeof prefix) prefix = '?';

  for (key in obj) {
    if (has.call(obj, key)) {
      value = obj[key];

      //
      // Edge cases where we actually want to encode the value to an empty
      // string instead of the stringified value.
      //
      if (!value && (value === null || value === undef || isNaN(value))) {
        value = '';
      }

      key = encodeURIComponent(key);
      value = encodeURIComponent(value);

      //
      // If we failed to encode the strings, we should bail out as we don't
      // want to add invalid strings to the query.
      //
      if (key === null || value === null) continue;
      pairs.push(key +'='+ value);
    }
  }

  return pairs.length ? prefix + pairs.join('&') : '';
}

/**
 * Simple cache for preloaded data
 */
var preloadCache = {};
var preloadJsonCache = {};
/**
 * An imperative API to preload data to implement render-as-you-fetch pattern
 */
var preloadData = function (client, url, queryParams, fetchOptions, dataOpts) {
    if (queryParams === void 0) { queryParams = {}; }
    if (fetchOptions === void 0) { fetchOptions = {}; }
    if (dataOpts === void 0) { dataOpts = {}; }
    var _a = createFetchOptions(fetchOptions, client, dataOpts), finalFetchOpts = _a[0], fetchPolicy = _a[1];
    var queryString = querystringify(queryParams, '?');
    var fullUrl = "" + url + queryString;
    var data = preloadCache[fullUrl];
    var fulfilled = data !== undefined;
    var error = null;
    var promise = !data
        ? fetch(fullUrl, finalFetchOpts)
            .then(function (result) { return result.text(); })
            .then(function (resText) {
            data = resText;
            if (fetchPolicy !== 'network-only') {
                preloadCache[fullUrl] = resText;
            }
            fulfilled = true;
        })
            .catch(function (e) { return (error = e); })
        : new Promise(function (resolve) {
            fulfilled = true;
            resolve();
        });
    var read = function () {
        if (error !== null) {
            // let an ErrorBoundary catch this
            throw error;
        }
        else if (!fulfilled && promise instanceof Promise) {
            // let a SuspenseBoundary catch this
            throw promise;
        }
        else {
            if (dataOpts.raw) {
                return data;
            }
            else {
                if (!preloadJsonCache[data]) {
                    preloadJsonCache[data] = JSON.parse(data);
                }
                return preloadJsonCache[data];
            }
        }
    };
    return {
        read: read,
    };
};

var createDataClient = function (options) {
    if (options === void 0) { options = {}; }
    var ssr = options.ssr, initialCache = options.initialCache, headers = options.headers, test = options.test, ssrForceFetchDelay = options.ssrForceFetchDelay, fetchPolicy = options.fetchPolicy;
    var subscribers = {};
    return {
        cache: initialCache ? __assign({}, initialCache) : {},
        ssr: ssr || false,
        ssrForceFetchDelay: ssrForceFetchDelay || 0,
        initTime: Date.now(),
        test: test || false,
        headers: headers || {},
        fetchPolicy: fetchPolicy || 'cache-first',
        toBePrefetched: {},
        addSubscriber: function (key, callback) {
            if (!(subscribers[key] instanceof Map)) {
                subscribers[key] = new Map();
            }
            subscribers[key].set(callback, callback);
        },
        removeSubscriber: function (key, callback) {
            subscribers[key].delete(callback);
        },
        notifySubscribers: function (key) {
            reactDom.unstable_batchedUpdates(function () {
                subscribers[key].forEach(function (callback) {
                    callback();
                });
            });
        },
    };
};

var areEqualShallow = function (a, b) {
    if (a === void 0) { a = {}; }
    if (b === void 0) { b = {}; }
    for (var key in a) {
        if (!(key in b) || a[key] !== b[key]) {
            return false;
        }
    }
    for (var key in b) {
        if (!(key in a) || a[key] !== b[key]) {
            return false;
        }
    }
    return true;
};

// returns an array of FetchOptions and string of fetchPolicy
var useFetchRequirements = function (fetchOptions, client, dataOpts, lazy) {
    if (lazy === void 0) { lazy = false; }
    var prevFetchRequirements = React.useRef();
    var prevFetchOpts = React.useRef();
    var prevDataOpts = React.useRef();
    var newFetchRequirements = createFetchOptions(fetchOptions, client, dataOpts, lazy);
    if (!prevFetchRequirements.current ||
        !areEqualShallow(fetchOptions, prevFetchOpts.current) ||
        !areEqualShallow(dataOpts, prevDataOpts.current)) {
        // we need to use new fetchRequirements
        prevFetchRequirements.current = newFetchRequirements;
    }
    prevFetchOpts.current = fetchOptions;
    prevDataOpts.current = dataOpts;
    return prevFetchRequirements.current;
};

var useIsomorphicLayoutEffect = typeof window !== 'undefined' &&
    typeof window.document !== 'undefined' &&
    typeof window.document.createElement !== 'undefined'
    ? React.useLayoutEffect
    : React.useEffect;

var LoadingSymbol = Symbol('LoadingFlag');
var subscribe = function (key, callback, client) {
    var listen = function () {
        // add the forceUpdate function to the cache's list of subscribers
        client.addSubscriber(key, callback);
    };
    var unlisten = function () {
        // remove from cache's list of subscriber
        client.removeSubscriber(key, callback);
    };
    return {
        listen: listen,
        unlisten: unlisten,
    };
};
/**
 * Create a subscription for the `string` key in the DataClient.cache
 */
var useCacheSubscription = function (key, lazy, skip) {
    var context = React.useContext(DataContext);
    if (!context)
        throw new Error('DataContext is null. Make sure you are wrapping your app inside DataProvider');
    var client = context.client, addToCache = context.addToCache, addToBePrefetched = context.addToBePrefetched, retrieveFromCache = context.retrieveFromCache, fetcher = context.fetcher;
    var _a = React.useState(0), setState = _a[1];
    var dataFromCache = retrieveFromCache(key);
    var initialLoading = lazy || skip ? false : true;
    if (dataFromCache && dataFromCache !== LoadingSymbol) {
        initialLoading = false;
    }
    var _b = React.useState({
        error: {},
        loading: initialLoading,
        tempCache: {},
    }), dataHookState = _b[0], setDataHookState = _b[1];
    var forceUpdate = React.useCallback(function () {
        setState(function (prev) { return prev + 1; });
    }, []);
    useIsomorphicLayoutEffect(function () {
        var subscription = subscribe(key, forceUpdate, client);
        subscription.listen();
        return function () {
            if (subscription)
                subscription.unlisten();
            subscription = null;
        };
    }, [key, forceUpdate, client]);
    return {
        client: client,
        addToCache: addToCache,
        addToBePrefetched: addToBePrefetched,
        retrieveFromCache: retrieveFromCache,
        dataFromCache: dataFromCache,
        dataHookState: dataHookState,
        setDataHookState: setDataHookState,
        fetcher: fetcher,
    };
};

var simpleCache = {};
var useBaseData = function (url, queryParams, fetchOptions, lazy, dataOpts) {
    if (queryParams === void 0) { queryParams = {}; }
    if (fetchOptions === void 0) { fetchOptions = {}; }
    if (lazy === void 0) { lazy = false; }
    if (dataOpts === void 0) { dataOpts = {}; }
    var _a;
    var queryString = querystringify(queryParams, '?');
    var fullUrl = "" + url + queryString;
    var ssrOpt = dataOpts.ssr !== undefined ? dataOpts.ssr : true;
    var skip = dataOpts.skip !== undefined ? dataOpts.skip : false;
    var raw = dataOpts.raw !== undefined ? dataOpts.raw : false;
    var _b = useCacheSubscription(fullUrl, lazy, skip), client = _b.client, addToCache = _b.addToCache, addToBePrefetched = _b.addToBePrefetched, retrieveFromCache = _b.retrieveFromCache, fetcher = _b.fetcher, dataFromCache = _b.dataFromCache, state = _b.dataHookState, setState = _b.setDataHookState;
    var _c = useFetchRequirements(fetchOptions, client, dataOpts, lazy), finalFetchOpts = _c[0], _fetchPolicy = _c[1];
    // add `<link rel="prefetch" /> tag for the resource only if it's enabled by user and the query isn't fetched during ssr
    var shouldPrefetch = dataOpts.prefetch !== undefined ? dataOpts.prefetch && (!ssrOpt || lazy) : false;
    var promiseRef = React.useRef(null);
    var promisePushed = React.useRef(false);
    var fetchedFromNetwork = React.useRef(false);
    var isSSR = client.ssr && ssrOpt && typeof window === 'undefined';
    var fetchPolicy = _fetchPolicy;
    if (!isSSR && _fetchPolicy === 'network-only' && client.ssrForceFetchDelay) {
        var delay = Date.now() - Number(client.initTime);
        if (delay <= client.ssrForceFetchDelay) {
            console.log('set to cache-first');
            fetchPolicy = 'cache-first'; // force to 'cache-first' policy when using ssrForceFetchDelay
        }
    }
    var useTempData = finalFetchOpts.method !== 'GET' || fetchPolicy === 'network-only';
    var createFetch = React.useCallback(function () {
        promiseRef.current = fetcher(fullUrl, finalFetchOpts)
            .then(function (result) { return result.text(); })
            .then(function (data) {
            // this block of code will cause 2 re-renders because React doesn't batch these 2 updates
            // https://twitter.com/dan_abramov/status/887963264335872000?lang=en
            // For React 16.x we can use `unstable_batchedUpdates()` to solve this
            reactDom.unstable_batchedUpdates(function () {
                if (!useTempData) {
                    // only cache response for GET requests
                    // AND non 'network-only' requests
                    addToCache(fullUrl, data);
                }
                else {
                    // resets the cache to 'undefined'
                    addToCache(fullUrl, undefined);
                }
                if (!isSSR) {
                    setState(function (prev) {
                        var _a, _b;
                        return (__assign(__assign({}, prev), { tempCache: __assign(__assign({}, prev.tempCache), (_a = {}, _a[fullUrl] = data, _a)), error: __assign(__assign({}, prev.error), (_b = {}, _b[fullUrl] = undefined, _b)) }));
                    });
                }
            });
            return data;
        })
            .catch(function (err) {
            if (!isSSR) {
                // sets the state accordingly
                reactDom.unstable_batchedUpdates(function () {
                    setState(function (prev) {
                        var _a;
                        return (__assign(__assign({}, prev), { error: __assign(__assign({}, prev.error), (_a = {}, _a[fullUrl] = err, _a)) }));
                    });
                    // resets the cache to 'undefined'
                    addToCache(fullUrl, undefined);
                });
            }
            else {
                // throw an error during SSR
                throw err;
            }
        });
        return promiseRef.current;
    }, [addToCache, finalFetchOpts, fullUrl, isSSR, useTempData, fetcher, setState]);
    var memoizedFetchData = React.useCallback(function () {
        var currentDataInCache = retrieveFromCache(fullUrl);
        // data not in cache yet
        if (currentDataInCache === undefined && !state.tempCache[fullUrl]) {
            addToCache(fullUrl, LoadingSymbol); // Use the loading flag as value temporarily
            fetchedFromNetwork.current = true;
            return createFetch();
        }
        // data is already in cache
        if (fetchPolicy !== 'cache-first') {
            // fetch again 1 time for cache-and-network cases
            // not that we do not set the cache to LoadingSymbol here,
            // this is so the component can reuse previous response while the new request is in progress
            if (!fetchedFromNetwork.current || lazy) {
                fetchedFromNetwork.current = true;
                return createFetch();
            }
        }
        return new Promise(function (resolve) { return resolve(currentDataInCache || state.tempCache[fullUrl]); });
    }, [
        lazy,
        state.tempCache,
        createFetch,
        fullUrl,
        addToCache,
        fetchPolicy,
        retrieveFromCache,
    ]);
    // if this data is supposed to be fetched during SSR
    if (isSSR) {
        if (!promisePushed.current && !lazy && !dataFromCache) {
            promisePushed.current = true;
            // throw a promise here. react-ssr-prepass will handle the suspension
            throw memoizedFetchData();
        }
    }
    // if the DataClient instance we are using is in ssr mode
    if (client.ssr) {
        if (shouldPrefetch) {
            addToBePrefetched(fullUrl);
        }
    }
    React.useEffect(function () {
        var _a;
        if (!lazy && !skip) {
            // !promiseRef.current ensure that the fetch is at least fired once.
            if ((dataFromCache !== LoadingSymbol && !((_a = state.error) === null || _a === void 0 ? void 0 : _a[fullUrl])) || !promiseRef.current) {
                memoizedFetchData();
            }
        }
    }, [skip, lazy, memoizedFetchData, dataFromCache, state.error, fullUrl]);
    var isLoading = dataFromCache === LoadingSymbol || (client.ssr && typeof dataFromCache === 'undefined' && !lazy);
    var finalData = dataFromCache !== LoadingSymbol ? dataFromCache : null;
    var usedData = (!useTempData ? finalData : state.tempCache[fullUrl]) || null;
    var memoizedData = React.useMemo(function () {
        /**
         * we only store the raw response string in the cache to avoid including double the amount of data
         * for the client.cache rehydration on client side
         * The caveats are:
         * 1. each hooks will need to do JSON.parse again to get the JSON
         *    So, we do some simple caching here to improve this
         * 2. Every re-renders means another string comparison
         *    Seems like string comparison isn't that expensive, even with long strings.
         *    So, let's settle with this for now
         *    (https://jsperf.com/eqaulity-is-constant-time)
         */
        if (raw) {
            // if we just want the raw data, just return it immediately
            return usedData;
        }
        else {
            // else, we want the data in json form
            if (!simpleCache[usedData]) {
                simpleCache[usedData] = JSON.parse(usedData);
            }
            return simpleCache[usedData];
        }
    }, [usedData, raw]);
    return [
        memoizedFetchData,
        {
            error: ((_a = state.error) === null || _a === void 0 ? void 0 : _a[fullUrl]) || null,
            loading: isLoading,
            data: memoizedData,
            refetch: function () {
                // always reset cache on refetch
                addToCache(fullUrl, LoadingSymbol); // Use the loading flag as value temporarily
                return createFetch();
            },
        },
    ];
};

var useData = function (url, queryParams, fetchOptions, dataOptions) {
    if (queryParams === void 0) { queryParams = {}; }
    if (fetchOptions === void 0) { fetchOptions = {}; }
    var _a = useBaseData(url, queryParams, fetchOptions, false, dataOptions), baseData = _a[1];
    return baseData;
};

var useData$1 = function (url, queryParams, fetchOptions, dataOptions) {
    if (queryParams === void 0) { queryParams = {}; }
    if (fetchOptions === void 0) { fetchOptions = {}; }
    var _a = useBaseData(url, queryParams, fetchOptions, true, dataOptions), fetchData = _a[0], baseData = _a[1];
    return [fetchData, baseData];
};

var useDataClient = function () {
    var context = React.useContext(DataContext);
    if (!context)
        throw new Error('DataContext is null. Make sure you are wrapping your app inside DataProvider');
    return context.client;
};

var withData = function (options) {
    var url = options.url, name = options.name, _a = options.queryParams, queryParams = _a === void 0 ? {} : _a, _b = options.fetchOptions, fetchOptions = _b === void 0 ? {} : _b, dataOptions = options.dataOptions;
    if (!name && "production" !== 'production') {
        console.warn('No `name` is passed to `withData` HOC, defaulting to `data`. Please provide a name to avoid props name collision!');
    }
    var Wrapped = function (Component) {
        var NewComp = function (props) {
            var _a;
            var baseData = useData(url, queryParams, fetchOptions, dataOptions);
            var dataProps = (_a = {},
                _a[name || 'data'] = baseData,
                _a);
            return React.createElement(Component, __assign({}, props, dataProps));
        };
        NewComp.displayName = "withData(" + (Component.displayName || Component.name || 'Component') + ")";
        return hoistNonReactStatics(React.memo(NewComp), Component);
    };
    return Wrapped;
};

var withData$1 = function (options) {
    var url = options.url, name = options.name, _a = options.queryParams, queryParams = _a === void 0 ? {} : _a, _b = options.fetchOptions, fetchOptions = _b === void 0 ? {} : _b, dataOptions = options.dataOptions;
    if (!name && "production" !== 'production') {
        console.warn('No `name` is passed to `withLazyData` HOC, defaulting to `data`. Please provide a name to avoid props name collision!');
    }
    var Wrapped = function (Component) {
        var NewComp = function (props) {
            var _a;
            var _b = useData$1(url, queryParams, fetchOptions, dataOptions), load = _b[0], baseData = _b[1];
            var dataProps = (_a = {},
                _a[name || 'data'] = [
                    load,
                    baseData
                ],
                _a);
            return React.createElement(Component, __assign({}, props, dataProps));
        };
        NewComp.displayName = "withLazyData(" + (Component.displayName || Component.name || 'Component') + ")";
        return hoistNonReactStatics(NewComp, Component);
    };
    return Wrapped;
};

var withDataClient = function (_a) {
    var _b = _a.name, name = _b === void 0 ? '' : _b;
    if (!name && "production" !== 'production') {
        console.warn('No `name` is passed to `withDataClient` HOC, defaulting to `client`. Please provide a name to avoid props name collision!');
    }
    return function (Component) {
        return hoistNonReactStatics(function (props) {
            var _a;
            var client = useDataClient();
            var dataProps = (_a = {},
                _a[name || 'client'] = client,
                _a);
            return React.createElement(Component, __assign({}, props, dataProps));
        }, Component);
    };
};

exports.DataContext = DataContext;
exports.DataProvider = DataProvider;
exports.addToCache = addToCache;
exports.createDataClient = createDataClient;
exports.preloadData = preloadData;
exports.retrieveFromCache = retrieveFromCache;
exports.useData = useData;
exports.useDataClient = useDataClient;
exports.useLazyData = useData$1;
exports.withData = withData;
exports.withDataClient = withDataClient;
exports.withLazyData = withData$1;
//# sourceMappingURL=index.js.map
